# Ball Tree wrapper for Fortran module.

# TODO: Investigate the correctness of the partial tree traversal
#       (budget < size(order)) when there are duplicate data points.

import os 
import numpy as np

# Allow OpenMP to create nested threads.
from multiprocessing import cpu_count

# Set OpenMP variables to allow for greatest parallelism when building tree.
if "OMP_NUM_THREADS" not in os.environ:
    os.environ["OMP_NUM_THREADS"] = str(cpu_count())
if "OMP_MAX_ACTIVE_LEVELS" not in os.environ:
    os.environ["OMP_MAX_ACTIVE_LEVELS"] = str(int(np.ceil(np.log2(cpu_count()))))
if "OMP_NESTED" not in os.environ:
    os.environ["OMP_NESTED"] = "TRUE"

# Import the Fortran utilities.
try:
    from .ball_tree import ball_tree
    from .prune import prune
    from .fast_sort import fast_sort
    from .fast_select import fast_select
except:
    print("Importing 'fmodpy' and building wrappers..")
    import fmodpy
    _PATH_TO_HERE = os.path.dirname(os.path.abspath(__file__))
    # Define paths to the Fortran utilities.
    _PATH_TO_BT = os.path.join(_PATH_TO_HERE, "ball_tree.f90")
    _PATH_TO_PRUNE = os.path.join(_PATH_TO_HERE, "prune.f90")
    _PATH_TO_SORT   = os.path.join(_PATH_TO_HERE, "fast_sort.f90")
    _PATH_TO_SELECT = os.path.join(_PATH_TO_HERE, "fast_select.f90")
    _DEPENDENCIES = ["swap.f90", "prune.f90", "fast_select.f90", "fast_sort.f90", "ball_tree.f90"]
    ball_tree  = fmodpy.fimport(_PATH_TO_BT, output_dir=_PATH_TO_HERE, omp=True,
                                verbose=False, dependencies=_DEPENDENCIES).ball_tree
    prune = fmodpy.fimport(_PATH_TO_PRUNE, output_dir=_PATH_TO_HERE, verbose=False).prune
    fast_sort = fmodpy.fimport(_PATH_TO_SORT, output_dir=_PATH_TO_HERE, verbose=False).fast_sort
    fast_select = fmodpy.fimport(_PATH_TO_SELECT, output_dir=_PATH_TO_HERE, verbose=False).fast_select

# ------------------------------------------------------------------
#                        FastSort method
# 
# This routine uses a combination of QuickSort (with modestly
# intelligent pivot selection) and Insertion Sort (for small arrays)
# to achieve very fast average case sort times for both random and
# partially sorted data. The pivot is selected for QuickSort as the
# median of the first, middle, and last values in the array.
# 
# Arguments:
# 
#   VALUES   --  A 1D array of real numbers.
# 
# Optional:
# 
#   INDICES  --  A 1D array of original indices for elements of VALUES.
#   MIN_SIZE --  An positive integer that represents the largest
#                sized VALUES for which a partition about a pivot
#                is used to reduce the size of a an unsorted array.
#                Any size less than this will result in the use of
#                INSERTION_ARGSORT instead of ARGPARTITION.
# 
# Output:
# 
#   The elements of the array VALUES are sorted and all elements of
#   INDICES are sorted symmetrically (given INDICES = 1, ...,
#   SIZE(VALUES) beforehand, final INDICES will show original index
#   of each element of VALUES before the sort operation).
# 
def argsort(values, indices=None, min_size=None):
    indices = np.arange(len(values))
    return fast_sort.argsort(values, indices, min_size=min_size)


# ------------------------------------------------------------------
#                       FastSelect method
# 
# Given VALUES list of numbers, rearrange the elements of VALUES
# such that the element at index K has rank K (holds its same
# location as if all of VALUES were sorted). Symmetrically rearrange
# array INDICES to keep track of prior indices.
# 
# This algorithm uses the same conceptual approach as Floyd-Rivest,
# but instead of standard-deviation based selection of bounds for
# recursion, a rank-based method is used to pick the subset of
# values that is searched. This simplifies the code and improves
# readability, while achieving the same tunable performance.
# 
# Arguments:
# 
#   VALUES   --  A 1D array of real numbers.
#   K        --  A positive integer for the rank index about which
#                VALUES should be rearranged.
# Optional:
# 
#   INDICES  --  A 1D array of original indices for elements of VALUES.
#   DIVISOR  --  A positive integer >= 2 that represents the
#                division factor used for large VALUES arrays.
#   MAX_SIZE --  An integer >= DIVISOR that represents the largest
#                sized VALUES for which the worst-case pivot value
#                selection is tolerable. A worst-case pivot causes
#                O( SIZE(VALUES)^2 ) runtime. This value should be
#                determined heuristically based on compute hardware.
# 
# Output:
# 
#   The elements of the array VALUES are rearranged such that the
#   element at position VALUES(K) is in the same location it would
#   be if all of VALUES were in sorted order. Also known as,
#   VALUES(K) has rank K.
# 
def argselect(values, k, indices=None, divisor=None, max_size=None):
    if (indices is None): indices = np.arange(len(values))
    return fast_select.argselect(values, indices, k+1, divisor=divisor, max_size=max_size)


# Class for constructing a ball tree.
class BallTree:
    # Given points and a leaf size, construct a ball tree.
    def __init__(self, points=None, dtype=None, transpose=True,
                 leaf_size=1, reorder=True, root=None):
        self.leaf_size = leaf_size
        # Assign the data type if it was provided.
        if (dtype is not None):
            self.ttype = dtype
            self._set_type_internals()
        else: self.ttype = None
        # Assign the various pruning functions.
        self._inner = prune.inner
        self._outer = prune.outer
        self._top = prune.top
        # Assign the points and build the tree, if provided.
        if (points is not None):
            self.add(points, transpose=transpose)
            self.build(reorder=reorder, root=root)

    # Given the data type of this class, setup which internal methods
    # are called (for the appropriate number sizes).
    def _set_type_internals(self):
        if ('float32' in str(self.ttype)):
            self.sstype = np.float32
            self._build_tree = ball_tree.build_tree
            self._fix_order  = ball_tree.fix_order
            self._bt_nearest = ball_tree.nearest
        else:
            class UnsupportedType(Exception): pass
            raise(UnsupportedType(f"The type '{self.ttype.name}' is not supported."))

    # Based on the size and 'built' amount and the 'size' internal: if
    # searching, return 'built' portion of tree; if building return
    # the 'size' portion of tree.
    def _get_order(self, search=False):
        # For a search, only use the built portion of the tree.
        if search:
            if (self.built == self.tree.shape[1]): return self.order
            else: return self.order[:self.built]
        # For a build, only use the 'size' kept points in the tree.
        else:
            if (self.size == self.tree.shape[1]): return self.order
            else: return self.order[:self.size]

    # Add points to this ball tree, if the type is not yet defined,
    # then initialize the type of this tree to be same as added points.
    def add(self, points, transpose=True):
        # Assign the data type if it is not set.
        if (self.ttype is None):
            self.ttype = np.float32
            self._set_type_internals()
        elif (points.dtype != self.ttype):
            class WrongType(Exception): pass
            raise(WrongType(f"The points provided were type '{points.dtype.name}', expected '{self.ttype.name}'."))
        # Transpose the points if the are not already column-vectors.
        if transpose: points = points.T
        # If there are existing points, make sure the new ones match.
        if hasattr(self, "size"):
            # If possible, pack the points into available space.
            if (points.shape[1] <= self.tree.shape[1] - self.size):
                self.tree[:,self.size:self.size+point.shape[1]] = points
                self.size += points.shape[1]
            else:
                old_tree = self.tree
                old_points = self.tree[:,:self.size]
                dim = self.tree.shape[1]
                self.size += points.shape[1]
                self.tree = np.zeros((dim, self.size), dtype=self.ttype, order='F')
                # Assign the relevant internals for this tree.
                self.sq_sums = np.zeros(self.tree.shape[1],  dtype=self.sstype)
                self.order   = np.arange(self.tree.shape[1], dtype='int64') + 1
                self.radii   = np.zeros(self.tree.shape[1],  dtype='float32')
                self.splits  = np.zeros(self.tree.shape[1],  dtype='float32')
                # Pack the old points and the new points into a single tree.
                self.tree[:,:old_points.shape[1]] = old_points
                self.tree[:,old_points.shape[1]:] = points
                # Delete the old tree.
                del old_tree
                self.built = 0
                return
        else:
            # Save the points internally as the tree.
            self.tree    = np.asarray(points, order='F', dtype=self.ttype)
            # Assign the relevant internals for this tree.
            self.sq_sums = np.zeros(self.tree.shape[1],  dtype=self.sstype)
            self.order   = np.arange(1, self.tree.shape[1]+1, dtype='int64')
            self.radii   = np.zeros(self.tree.shape[1],  dtype='float32')
            self.splits  = np.zeros(self.tree.shape[1],  dtype='float32')
            # Store BallTree internals for knowing how to evaluate.
            self.built = 0
            self.size = self.tree.shape[1]

    # Build a tree out.
    def build(self, leaf_size=None, reorder=True, has_sq_sums=False, root=None):
        # Get the leaf size if it was not given.
        if (leaf_size is None): leaf_size = self.leaf_size
        # Translate the root from python index to fortran index if provided.
        if (root is not None): root += 1
        # Automatically set 'reorder' to False if the array is more than
        # 2 GB in size because this requires a doubling of memory usage.
        if (reorder is None): reorder = (self.tree.nbytes / 2**30) <= 2
        # Build tree (in-place operation).
        #    .tree     will not be modified
        #    .sq_sums  will contain the squared sums of each point
        #    .radii    will be modified to have the radius of specific
        #              node, or 0.0 if this point is a leaf.
        #    .splits   will be modified to have the distance to the "split"
        #              median child node, or 0.0 if this point is a leaf.
        #    .order    will be the list of indices (1-indexed) that
        #              determine the structure of the ball tree.
        order = self._get_order(search=False)
        self._build_tree(self.tree, self.sq_sums, self.radii, self.splits, order,
                         leaf_size=leaf_size, computed_sq_sums=has_sq_sums, 
                         root=root)
        self.built = len(order)
        # Store the index mapping from the build.
        if hasattr(self, "index_mapping"): del self.index_mapping
        self.index_mapping = order.copy() - 1
        # Restructure the ball tree so the points are in locally
        # contiguous blocks of memory (local by branch + leaf).
        if reorder: self._fix_order(self.tree, self.sq_sums, self.radii, self.splits, order)

    # Restructure the ball tree so the points are in locally
    # contiguous blocks of memory (local by branch + leaf).
    def reorder(self):
        order = self._get_order(search=True)
        self._fix_order(self.tree, self.sq_sums, self.radii, self.splits, order)

    # Find the "k" nearest neighbors to all points in z. Uses the same
    # interface as the "BallTree.nearest" function, see help for more info.
    def query(self, *args, **kwargs): return self.nearest(*args, **kwargs)

    # Find the "k" nearest neighbors to all points in z.
    def nearest(self, z, k=1, leaf_size=None, return_distance=True,
                transpose=True, budget=None, randomness=None):
        # Get the leaf size.
        if (leaf_size is None): leaf_size = self.leaf_size
        # If only a single point was given, convert it to a matrix.
        if (len(z.shape) == 1): z = np.array([z])
        # Transpose the points if appropriate.
        if transpose: z = z.T
        # Make sure the 'k' value isn't bigger than the tree size.
        k = min(k, self.built)
        # Initialize holders for output.
        points  = np.asarray(z, order='F', dtype=self.ttype)
        indices = np.ones((k, points.shape[1]), order='F', dtype='int64')
        dists   = np.ones((k, points.shape[1]), order='F', dtype='float32')
        order = self._get_order(search=True)
        # Compute the nearest neighbors.
        self._bt_nearest(points, k, self.tree, self.sq_sums, self.radii,
                         self.splits, order, leaf_size, indices, dists,
                         to_search=budget, randomness=randomness)
        # Return the results.
        if return_distance:
            if transpose: return dists.T, indices.T - 1
            else:         return dists,   indices - 1
        else:
            if transpose: return indices.T - 1
            else:         return indices - 1

    # Summarize this tree.
    def __str__(self):
        if not hasattr(self,"tree"): return "empty BallTree"
        return f"BallTree {self.tree.shape[::-1]}[:{self.size}] -- {self.built} built"

    # Return the usable length of this tree.
    def __len__(self):
        if not hasattr(self,"built"): return 0
        else:                         return self.built

    # Get an index point from the tree.
    def __getitem__(self, index):
        return self.tree[:,self.order[:self.built][index]-1]

    # Prune this tree and compact its points into the front of the
    # array, adjust '.size' and '.built' accordingly.
    def prune(self, levels=1, full=True, build=True, method="root"):
        assert(levels >= 1)
        assert(len(self) > 0)
        # Build the tree out if it is not fully built.
        if (full and (not (self.built == self.size))): self.build()
        # Get the size of the built portion of the tree.
        size = self.built
        # Compute the indices of the inner children (50th percentile) points.
        if (levels == 1):
            # Use the middle child for 1 level.
            indices = np.array([1], dtype=np.int64)
        else:
            # Handle the different methods of pruning the tree.
            if method in {"inner", "outer"}:
                to_keep = min(size, 1 + 2**(levels-1))
                # Otherwise, get the root, first outer, and all inners.
                indices = np.zeros(to_keep, dtype=np.int64)
                # Get the indices of the inner children of the built tree.
                if method == "inner":
                    indices[0] = 1
                    indices[1] = min((size + 2) // 2, size-1) + 1
                    self._inner(size, levels, indices[2:])
                elif method == "outer":
                    indices[0] = 1
                    indices[1] = 2
                    self._outer(size, levels, indices[2:])
            else:
                to_keep = min(size, 2**levels - 1)
                # Simply grab the root of the tree.
                indices = np.zeros(to_keep, dtype=np.int64)
                self._top(size, levels, indices)
            indices[:] -= 1
        # Stop this operation if the tree will remain unchanged.
        if (len(indices) == size): return
        # Adjust the recorded size and amount built of this tree.
        self.size = len(indices)
        self.built = 1
        # Keep those indices only.
        self.order[:self.size] = self.order[indices]
        # Rebuild the tree if desired.
        if build: self.build(root=0)




import numpy as np
import fmodpy

# Matrix operations module.
mops = fmodpy.fimport("../matrix_operations.f90", blas=True, lapack=True).matrix_operations

# --------------------------------------------------------------------
#                         ORTHOGONALIZE

# Test orthogonalization.
d = 4
m = np.random.random(size=(d,d)).astype("float32")
l = np.zeros(d, dtype="float32")

print("m: ")
print(m)
print(m.T @ m)
print(m @ m.T)
print()

mops.orthogonalize(m.T, l)

print("m: ")
print(m)
print(m.T @ m)
print(m @ m.T)
print()

print("l: ")
print(l)
print()

exit()

# --------------------------------------------------------------------
#                        RANDOM_UNIT_VECTORS

# Generate test data.
n = 100 # number of points
d = 2   # dimenion of points

bd = 2 # full number of outputs
br = 1 # reduced rank approximation

# Random unit vectors.
a = np.zeros((n,d), dtype="float32")
mops.random_unit_vectors(a.T)

# Generate target outptus.
x = np.zeros((bd,d), dtype="float32").T
mops.random_unit_vectors(x)
b = np.array(np.matmul(a, x), order="F", dtype="float32")

# --------------------------------------------------------------------
#                            LEAST_SQUARES

# Use the matrix operations subroutine to solve the least squares problem.
mo_x = np.zeros((d,br), order="F", dtype="float32")
mops.least_squares(
    bytes("T","ascii"),
    np.array(a.T, order="F", dtype="float32"),
    np.array(b, order="F", dtype="float32"),
    mo_x)

np_x, residual = np.linalg.lstsq(a, b, rcond=None)[:2]

print()
print("x: ")
print(x)

print()
print("np_x: ")
print(np_x)

print()
print("mo_x: ")
print(mo_x)

print()
print("mo_x (scaled to full output):")
upscale = np.linalg.lstsq(np.matmul(a, mo_x), b, rcond=None)[0]
print(np.matmul(mo_x, upscale))


# Generate a visiaul.
from tlux.plot import Plot
p = Plot()
p.add("a", *a.T, color=1)
# Show the true values.
for i in range(bd):
    p.add(f"b{i}", *a.T, b[:,i], color=2+i, shade=True)
# Show the literal values produced by the mo regression.
for i in range(br):
    p.add(f"app b{i}", *a.T, np.matmul(a, mo_x)[:,i],
          color=(0,0,0,0.8), marker_size=2)
# Show the original approximations recovered from the mo regression.
for i in range(bd):
    p.add(f"recon b{i}", *a.T, np.matmul(a, np.matmul(mo_x, upscale))[:,i],
          color=2+i, shade=True, marker_size=4, marker_line_width=1)

# Show the plot, different depending on if it is 2D (points only) or 3D (with regression).
if (bd > 0):
    p.show(x_range=[-1.1,1.1], y_range=[-1.1,1.1], z_range=[-1.1,1.1])
else:
    p.show(x_range=[-2.3,2.3], y_range=[-1.1,1.1])



'''This Python code is an automatically generated wrapper
for Fortran code made by 'fmodpy'. The original documentation
for the Fortran source code follows.


'''

import os
import ctypes
import platform
import numpy

# --------------------------------------------------------------------
#               CONFIGURATION
# 
_verbose = True
_fort_compiler = "gfortran"
_shared_object_name = "ball_tree." + platform.machine() + ".so"
_this_directory = os.path.dirname(os.path.abspath(__file__))
_path_to_lib = os.path.join(_this_directory, _shared_object_name)
_compile_options = ['-fPIC', '-shared', '-O3', '-fopenmp']
_ordered_dependencies = ['swap.f90', 'prune.f90', 'fast_select.f90', 'fast_sort.f90', 'ball_tree.f90', 'ball_tree_c_wrapper.f90']
# 
# --------------------------------------------------------------------
#               AUTO-COMPILING
#
# Try to import the existing object. If that fails, recompile and then try.
try:
    clib = ctypes.CDLL(_path_to_lib)
except:
    # Remove the shared object if it exists, because it is faulty.
    if os.path.exists(_shared_object_name):
        os.remove(_shared_object_name)
    # Compile a new shared object.
    _command = " ".join([_fort_compiler] + _compile_options + ["-o", _shared_object_name] + _ordered_dependencies)
    if _verbose:
        print("Running system command with arguments")
        print("  ", _command)
    # Run the compilation command.
    import subprocess
    subprocess.run(_command, shell=True, cwd=_this_directory)
    # Import the shared object file as a C library with ctypes.
    clib = ctypes.CDLL(_path_to_lib)
# --------------------------------------------------------------------


class ball_tree:
    ''''''

    # Declare 'max_copy_bytes'
    def get_max_copy_bytes(self):
        max_copy_bytes = ctypes.c_long()
        clib.ball_tree_get_max_copy_bytes(ctypes.byref(max_copy_bytes))
        return max_copy_bytes.value
    def set_max_copy_bytes(self, max_copy_bytes):
        max_copy_bytes = ctypes.c_long(max_copy_bytes)
        clib.ball_tree_set_max_copy_bytes(ctypes.byref(max_copy_bytes))
    max_copy_bytes = property(get_max_copy_bytes, set_max_copy_bytes)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine BUILD_TREE
    
    def build_tree(self, points, sq_sums, radii, splits, order, root=None, leaf_size=None, computed_sq_sums=None):
        '''! Re-arrange elements of POINTS into a binary ball tree.'''
        
        # Setting up "points"
        if ((not issubclass(type(points), numpy.ndarray)) or
            (not numpy.asarray(points).flags.f_contiguous) or
            (not (points.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'points' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            points = numpy.asarray(points, dtype=ctypes.c_float, order='F')
        points_dim_1 = ctypes.c_long(points.shape[0])
        points_dim_2 = ctypes.c_long(points.shape[1])
        
        # Setting up "sq_sums"
        if ((not issubclass(type(sq_sums), numpy.ndarray)) or
            (not numpy.asarray(sq_sums).flags.f_contiguous) or
            (not (sq_sums.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'sq_sums' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            sq_sums = numpy.asarray(sq_sums, dtype=ctypes.c_float, order='F')
        sq_sums_dim_1 = ctypes.c_long(sq_sums.shape[0])
        
        # Setting up "radii"
        if ((not issubclass(type(radii), numpy.ndarray)) or
            (not numpy.asarray(radii).flags.f_contiguous) or
            (not (radii.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'radii' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            radii = numpy.asarray(radii, dtype=ctypes.c_float, order='F')
        radii_dim_1 = ctypes.c_long(radii.shape[0])
        
        # Setting up "splits"
        if ((not issubclass(type(splits), numpy.ndarray)) or
            (not numpy.asarray(splits).flags.f_contiguous) or
            (not (splits.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'splits' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            splits = numpy.asarray(splits, dtype=ctypes.c_float, order='F')
        splits_dim_1 = ctypes.c_long(splits.shape[0])
        
        # Setting up "order"
        if ((not issubclass(type(order), numpy.ndarray)) or
            (not numpy.asarray(order).flags.f_contiguous) or
            (not (order.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'order' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            order = numpy.asarray(order, dtype=ctypes.c_long, order='F')
        order_dim_1 = ctypes.c_long(order.shape[0])
        
        # Setting up "root"
        root_present = ctypes.c_bool(True)
        if (root is None):
            root_present = ctypes.c_bool(False)
            root = ctypes.c_long()
        else:
            root = ctypes.c_long(root)
        if (type(root) is not ctypes.c_long): root = ctypes.c_long(root)
        
        # Setting up "leaf_size"
        leaf_size_present = ctypes.c_bool(True)
        if (leaf_size is None):
            leaf_size_present = ctypes.c_bool(False)
            leaf_size = ctypes.c_long()
        else:
            leaf_size = ctypes.c_long(leaf_size)
        if (type(leaf_size) is not ctypes.c_long): leaf_size = ctypes.c_long(leaf_size)
        
        # Setting up "computed_sq_sums"
        computed_sq_sums_present = ctypes.c_bool(True)
        if (computed_sq_sums is None):
            computed_sq_sums_present = ctypes.c_bool(False)
            computed_sq_sums = ctypes.c_int()
        else:
            computed_sq_sums = ctypes.c_int(computed_sq_sums)
        if (type(computed_sq_sums) is not ctypes.c_int): computed_sq_sums = ctypes.c_int(computed_sq_sums)
    
        # Call C-accessible Fortran wrapper.
        clib.c_build_tree(ctypes.byref(points_dim_1), ctypes.byref(points_dim_2), ctypes.c_void_p(points.ctypes.data), ctypes.byref(sq_sums_dim_1), ctypes.c_void_p(sq_sums.ctypes.data), ctypes.byref(radii_dim_1), ctypes.c_void_p(radii.ctypes.data), ctypes.byref(splits_dim_1), ctypes.c_void_p(splits.ctypes.data), ctypes.byref(order_dim_1), ctypes.c_void_p(order.ctypes.data), ctypes.byref(root_present), ctypes.byref(root), ctypes.byref(leaf_size_present), ctypes.byref(leaf_size), ctypes.byref(computed_sq_sums_present), ctypes.byref(computed_sq_sums))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return points, sq_sums, radii, splits, order

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine NEAREST
    
    def nearest(self, points, k, tree, sq_sums, radii, splits, order, leaf_size, indices=None, dists=None, to_search=None, randomness=None):
        '''! Compute the K nearest elements of TREE to each point in POINTS.'''
        
        # Setting up "points"
        if ((not issubclass(type(points), numpy.ndarray)) or
            (not numpy.asarray(points).flags.f_contiguous) or
            (not (points.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'points' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            points = numpy.asarray(points, dtype=ctypes.c_float, order='F')
        points_dim_1 = ctypes.c_long(points.shape[0])
        points_dim_2 = ctypes.c_long(points.shape[1])
        
        # Setting up "k"
        if (type(k) is not ctypes.c_long): k = ctypes.c_long(k)
        
        # Setting up "tree"
        if ((not issubclass(type(tree), numpy.ndarray)) or
            (not numpy.asarray(tree).flags.f_contiguous) or
            (not (tree.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'tree' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            tree = numpy.asarray(tree, dtype=ctypes.c_float, order='F')
        tree_dim_1 = ctypes.c_long(tree.shape[0])
        tree_dim_2 = ctypes.c_long(tree.shape[1])
        
        # Setting up "sq_sums"
        if ((not issubclass(type(sq_sums), numpy.ndarray)) or
            (not numpy.asarray(sq_sums).flags.f_contiguous) or
            (not (sq_sums.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'sq_sums' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            sq_sums = numpy.asarray(sq_sums, dtype=ctypes.c_float, order='F')
        sq_sums_dim_1 = ctypes.c_long(sq_sums.shape[0])
        
        # Setting up "radii"
        if ((not issubclass(type(radii), numpy.ndarray)) or
            (not numpy.asarray(radii).flags.f_contiguous) or
            (not (radii.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'radii' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            radii = numpy.asarray(radii, dtype=ctypes.c_float, order='F')
        radii_dim_1 = ctypes.c_long(radii.shape[0])
        
        # Setting up "splits"
        if ((not issubclass(type(splits), numpy.ndarray)) or
            (not numpy.asarray(splits).flags.f_contiguous) or
            (not (splits.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'splits' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            splits = numpy.asarray(splits, dtype=ctypes.c_float, order='F')
        splits_dim_1 = ctypes.c_long(splits.shape[0])
        
        # Setting up "order"
        if ((not issubclass(type(order), numpy.ndarray)) or
            (not numpy.asarray(order).flags.f_contiguous) or
            (not (order.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'order' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            order = numpy.asarray(order, dtype=ctypes.c_long, order='F')
        order_dim_1 = ctypes.c_long(order.shape[0])
        
        # Setting up "leaf_size"
        if (type(leaf_size) is not ctypes.c_long): leaf_size = ctypes.c_long(leaf_size)
        
        # Setting up "indices"
        if (indices is None):
            indices = numpy.zeros(shape=(k, points.shape[1]), dtype=ctypes.c_long, order='F')
        elif ((not issubclass(type(indices), numpy.ndarray)) or
              (not numpy.asarray(indices).flags.f_contiguous) or
              (not (indices.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'indices' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            indices = numpy.asarray(indices, dtype=ctypes.c_long, order='F')
        indices_dim_1 = ctypes.c_long(indices.shape[0])
        indices_dim_2 = ctypes.c_long(indices.shape[1])
        
        # Setting up "dists"
        if (dists is None):
            dists = numpy.zeros(shape=(k, points.shape[1]), dtype=ctypes.c_float, order='F')
        elif ((not issubclass(type(dists), numpy.ndarray)) or
              (not numpy.asarray(dists).flags.f_contiguous) or
              (not (dists.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'dists' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            dists = numpy.asarray(dists, dtype=ctypes.c_float, order='F')
        dists_dim_1 = ctypes.c_long(dists.shape[0])
        dists_dim_2 = ctypes.c_long(dists.shape[1])
        
        # Setting up "to_search"
        to_search_present = ctypes.c_bool(True)
        if (to_search is None):
            to_search_present = ctypes.c_bool(False)
            to_search = ctypes.c_long()
        else:
            to_search = ctypes.c_long(to_search)
        if (type(to_search) is not ctypes.c_long): to_search = ctypes.c_long(to_search)
        
        # Setting up "randomness"
        randomness_present = ctypes.c_bool(True)
        if (randomness is None):
            randomness_present = ctypes.c_bool(False)
            randomness = ctypes.c_float()
        else:
            randomness = ctypes.c_float(randomness)
        if (type(randomness) is not ctypes.c_float): randomness = ctypes.c_float(randomness)
    
        # Call C-accessible Fortran wrapper.
        clib.c_nearest(ctypes.byref(points_dim_1), ctypes.byref(points_dim_2), ctypes.c_void_p(points.ctypes.data), ctypes.byref(k), ctypes.byref(tree_dim_1), ctypes.byref(tree_dim_2), ctypes.c_void_p(tree.ctypes.data), ctypes.byref(sq_sums_dim_1), ctypes.c_void_p(sq_sums.ctypes.data), ctypes.byref(radii_dim_1), ctypes.c_void_p(radii.ctypes.data), ctypes.byref(splits_dim_1), ctypes.c_void_p(splits.ctypes.data), ctypes.byref(order_dim_1), ctypes.c_void_p(order.ctypes.data), ctypes.byref(leaf_size), ctypes.byref(indices_dim_1), ctypes.byref(indices_dim_2), ctypes.c_void_p(indices.ctypes.data), ctypes.byref(dists_dim_1), ctypes.byref(dists_dim_2), ctypes.c_void_p(dists.ctypes.data), ctypes.byref(to_search_present), ctypes.byref(to_search), ctypes.byref(randomness_present), ctypes.byref(randomness))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return indices, dists

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine PT_NEAREST
    
    def pt_nearest(self, point, k, tree, sq_sums, radii, splits, order, leaf_size, indices, dists, randomness, checks=None, found=None, pt_ss=None, d_root=None):
        '''! Compute the K nearest elements of TREE to each point in POINTS.'''
        
        # Setting up "point"
        if ((not issubclass(type(point), numpy.ndarray)) or
            (not numpy.asarray(point).flags.f_contiguous) or
            (not (point.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'point' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            point = numpy.asarray(point, dtype=ctypes.c_float, order='F')
        point_dim_1 = ctypes.c_long(point.shape[0])
        
        # Setting up "k"
        if (type(k) is not ctypes.c_long): k = ctypes.c_long(k)
        
        # Setting up "tree"
        if ((not issubclass(type(tree), numpy.ndarray)) or
            (not numpy.asarray(tree).flags.f_contiguous) or
            (not (tree.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'tree' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            tree = numpy.asarray(tree, dtype=ctypes.c_float, order='F')
        tree_dim_1 = ctypes.c_long(tree.shape[0])
        tree_dim_2 = ctypes.c_long(tree.shape[1])
        
        # Setting up "sq_sums"
        if ((not issubclass(type(sq_sums), numpy.ndarray)) or
            (not numpy.asarray(sq_sums).flags.f_contiguous) or
            (not (sq_sums.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'sq_sums' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            sq_sums = numpy.asarray(sq_sums, dtype=ctypes.c_float, order='F')
        sq_sums_dim_1 = ctypes.c_long(sq_sums.shape[0])
        
        # Setting up "radii"
        if ((not issubclass(type(radii), numpy.ndarray)) or
            (not numpy.asarray(radii).flags.f_contiguous) or
            (not (radii.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'radii' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            radii = numpy.asarray(radii, dtype=ctypes.c_float, order='F')
        radii_dim_1 = ctypes.c_long(radii.shape[0])
        
        # Setting up "splits"
        if ((not issubclass(type(splits), numpy.ndarray)) or
            (not numpy.asarray(splits).flags.f_contiguous) or
            (not (splits.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'splits' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            splits = numpy.asarray(splits, dtype=ctypes.c_float, order='F')
        splits_dim_1 = ctypes.c_long(splits.shape[0])
        
        # Setting up "order"
        if ((not issubclass(type(order), numpy.ndarray)) or
            (not numpy.asarray(order).flags.f_contiguous) or
            (not (order.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'order' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            order = numpy.asarray(order, dtype=ctypes.c_long, order='F')
        order_dim_1 = ctypes.c_long(order.shape[0])
        
        # Setting up "leaf_size"
        if (type(leaf_size) is not ctypes.c_long): leaf_size = ctypes.c_long(leaf_size)
        
        # Setting up "indices"
        if ((not issubclass(type(indices), numpy.ndarray)) or
            (not numpy.asarray(indices).flags.f_contiguous) or
            (not (indices.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'indices' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            indices = numpy.asarray(indices, dtype=ctypes.c_long, order='F')
        indices_dim_1 = ctypes.c_long(indices.shape[0])
        
        # Setting up "dists"
        if ((not issubclass(type(dists), numpy.ndarray)) or
            (not numpy.asarray(dists).flags.f_contiguous) or
            (not (dists.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'dists' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            dists = numpy.asarray(dists, dtype=ctypes.c_float, order='F')
        dists_dim_1 = ctypes.c_long(dists.shape[0])
        
        # Setting up "randomness"
        if (type(randomness) is not ctypes.c_float): randomness = ctypes.c_float(randomness)
        
        # Setting up "checks"
        checks_present = ctypes.c_bool(True)
        if (checks is None):
            checks_present = ctypes.c_bool(False)
            checks = ctypes.c_long()
        else:
            checks = ctypes.c_long(checks)
        if (type(checks) is not ctypes.c_long): checks = ctypes.c_long(checks)
        
        # Setting up "found"
        found_present = ctypes.c_bool(True)
        if (found is None):
            found_present = ctypes.c_bool(False)
            found = ctypes.c_long()
        else:
            found = ctypes.c_long(found)
        if (type(found) is not ctypes.c_long): found = ctypes.c_long(found)
        
        # Setting up "pt_ss"
        pt_ss_present = ctypes.c_bool(True)
        if (pt_ss is None):
            pt_ss_present = ctypes.c_bool(False)
            pt_ss = ctypes.c_float()
        else:
            pt_ss = ctypes.c_float(pt_ss)
        if (type(pt_ss) is not ctypes.c_float): pt_ss = ctypes.c_float(pt_ss)
        
        # Setting up "d_root"
        d_root_present = ctypes.c_bool(True)
        if (d_root is None):
            d_root_present = ctypes.c_bool(False)
            d_root = ctypes.c_float()
        else:
            d_root = ctypes.c_float(d_root)
        if (type(d_root) is not ctypes.c_float): d_root = ctypes.c_float(d_root)
    
        # Call C-accessible Fortran wrapper.
        clib.c_pt_nearest(ctypes.byref(point_dim_1), ctypes.c_void_p(point.ctypes.data), ctypes.byref(k), ctypes.byref(tree_dim_1), ctypes.byref(tree_dim_2), ctypes.c_void_p(tree.ctypes.data), ctypes.byref(sq_sums_dim_1), ctypes.c_void_p(sq_sums.ctypes.data), ctypes.byref(radii_dim_1), ctypes.c_void_p(radii.ctypes.data), ctypes.byref(splits_dim_1), ctypes.c_void_p(splits.ctypes.data), ctypes.byref(order_dim_1), ctypes.c_void_p(order.ctypes.data), ctypes.byref(leaf_size), ctypes.byref(indices_dim_1), ctypes.c_void_p(indices.ctypes.data), ctypes.byref(dists_dim_1), ctypes.c_void_p(dists.ctypes.data), ctypes.byref(randomness), ctypes.byref(checks_present), ctypes.byref(checks), ctypes.byref(found_present), ctypes.byref(found), ctypes.byref(pt_ss_present), ctypes.byref(pt_ss), ctypes.byref(d_root_present), ctypes.byref(d_root))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return indices, dists, (checks.value if checks_present else None), (found.value if found_present else None)

    
    # ----------------------------------------------
    # Wrapper for the Fortran subroutine FIX_ORDER
    
    def fix_order(self, points, sq_sums, radii, splits, order, copy=None):
        '''! Re-organize a built tree so that it is more usefully packed in memory.'''
        
        # Setting up "points"
        if ((not issubclass(type(points), numpy.ndarray)) or
            (not numpy.asarray(points).flags.f_contiguous) or
            (not (points.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'points' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            points = numpy.asarray(points, dtype=ctypes.c_float, order='F')
        points_dim_1 = ctypes.c_long(points.shape[0])
        points_dim_2 = ctypes.c_long(points.shape[1])
        
        # Setting up "sq_sums"
        if ((not issubclass(type(sq_sums), numpy.ndarray)) or
            (not numpy.asarray(sq_sums).flags.f_contiguous) or
            (not (sq_sums.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'sq_sums' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            sq_sums = numpy.asarray(sq_sums, dtype=ctypes.c_float, order='F')
        sq_sums_dim_1 = ctypes.c_long(sq_sums.shape[0])
        
        # Setting up "radii"
        if ((not issubclass(type(radii), numpy.ndarray)) or
            (not numpy.asarray(radii).flags.f_contiguous) or
            (not (radii.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'radii' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            radii = numpy.asarray(radii, dtype=ctypes.c_float, order='F')
        radii_dim_1 = ctypes.c_long(radii.shape[0])
        
        # Setting up "splits"
        if ((not issubclass(type(splits), numpy.ndarray)) or
            (not numpy.asarray(splits).flags.f_contiguous) or
            (not (splits.dtype == numpy.dtype(ctypes.c_float)))):
            import warnings
            warnings.warn("The provided argument 'splits' was not an f_contiguous NumPy array of type 'ctypes.c_float' (or equivalent). Automatically converting (probably creating a full copy).")
            splits = numpy.asarray(splits, dtype=ctypes.c_float, order='F')
        splits_dim_1 = ctypes.c_long(splits.shape[0])
        
        # Setting up "order"
        if ((not issubclass(type(order), numpy.ndarray)) or
            (not numpy.asarray(order).flags.f_contiguous) or
            (not (order.dtype == numpy.dtype(ctypes.c_long)))):
            import warnings
            warnings.warn("The provided argument 'order' was not an f_contiguous NumPy array of type 'ctypes.c_long' (or equivalent). Automatically converting (probably creating a full copy).")
            order = numpy.asarray(order, dtype=ctypes.c_long, order='F')
        order_dim_1 = ctypes.c_long(order.shape[0])
        
        # Setting up "copy"
        copy_present = ctypes.c_bool(True)
        if (copy is None):
            copy_present = ctypes.c_bool(False)
            copy = ctypes.c_int()
        else:
            copy = ctypes.c_int(copy)
        if (type(copy) is not ctypes.c_int): copy = ctypes.c_int(copy)
    
        # Call C-accessible Fortran wrapper.
        clib.c_fix_order(ctypes.byref(points_dim_1), ctypes.byref(points_dim_2), ctypes.c_void_p(points.ctypes.data), ctypes.byref(sq_sums_dim_1), ctypes.c_void_p(sq_sums.ctypes.data), ctypes.byref(radii_dim_1), ctypes.c_void_p(radii.ctypes.data), ctypes.byref(splits_dim_1), ctypes.c_void_p(splits.ctypes.data), ctypes.byref(order_dim_1), ctypes.c_void_p(order.ctypes.data), ctypes.byref(copy_present), ctypes.byref(copy))
    
        # Return final results, 'INTENT(OUT)' arguments only.
        return points, sq_sums, radii, splits, order

ball_tree = ball_tree()

